#!/usr/bin/perl -I/usr/sausalito/perl
#
# $Id: vzbackup-vzdump-hook.pl, v 2.0.0-1 Sun 30 Oct 2011 10:42:01 AM CET mstauber Exp $
# Copyright 2006-2011 Stauber Multimedia Design. All rights reserved.
#
# This is a vzdump hook script. It is used by vzdump to update CODB with information about
# the performed backup transactions. It also performs the transfer of the remote backup to
# its intended destination. If need be (if local backup is disabled) it will also delete
# the local tarball and logfile.

use CCE;
use POSIX qw(isalpha);
use Socket;
use Sys::Hostname;
use File::Spec;
use List::MoreUtils qw(uniq);
use Net::SFTP::Foreign;
use File::Basename;
use POSIX qw(strftime);

my $cce = new CCE;
$cce->connectuds();

$oid = $cce->event_oid();
$obj = $cce->event_object();

$host = hostname();
$addr = inet_ntoa(scalar(gethostbyname($host)) || 'localhost');
$now = strftime('%Y-%m-%d %H:%M:%S', localtime);

my $phase = shift;
my $mode = shift; # stop/suspend/snapshot
my $vmid = shift;
my $vmtype = $ENV{VMTYPE}; # openvz/qemu
my $dumpdir = $ENV{DUMPDIR};
my $hostname = $ENV{HOSTNAME};
my $datetime = $now;

# tarfile is only available in phase 'backup-end'
my $tarfile = $ENV{TARFILE};

# logfile is only available in phase 'log-end'
my $logfile = $ENV{LOGFILE};

# Helper to redirect SSH errors to /dev/null:
open $ssherr, '>', '/dev/null'; # or die "Unable to open /dev/null";

# Get settings out of CODB to determine the target of the remote backup (if any):
@oids = $cce->find('VserverMasterNode');
($ok, $solbackupValues) = $cce->get($oids[0]);
if ($ok) {
    # Define values:
    $backupdir              	= $solbackupValues->{'MN_BackupLocalLocation'} . "/vzbackup";
    $dumpdir                	= $solbackupValues->{'MN_BackupLocalLocation'} . "/vzdump";
    $targethost 		= $solbackupValues->{'MN_BackupRemoteTargetHost'};
    $targetip 			= $solbackupValues->{'MN_BackupRemoteTargetIP'};
    $path 			= $solbackupValues->{'MN_BackupRemoteLocation'};
    $path 			=~ s|//|/|g;
    $path 			=~ s|/$||g;
    $user 			= $solbackupValues->{'MN_BackupRemoteTargetUser'};
    $pass 			= $solbackupValues->{'MN_BackupRemoteTargetUserPass'};
    $port 			= $solbackupValues->{'MN_BackupRemoteTargetPort'};
    $limit 			= $solbackupValues->{'RsyncSpeedLimit'};
    $my_maxfiles 		= $solbackupValues->{'vzdump_maxfiles'};
}


# Find out of this VPS has remote backup enabled:
($sys_oid) = $cce->find('Vserver_Backup', {'vpsid' => $vmid});
($ok, $vpsinfo) = $cce->get($sys_oid);
if ($ok) {
    $remote_on		= $vpsinfo->{'VserverBackupRemoteEnabled'};
    $local_on		= $vpsinfo->{'VserverBackupLocalEnabled'};
    $virttype		= $vpsinfo->{'virttype'};
}

# Get IP and FQDN of the VPS:
if ($virttype eq "openvz") {
    ($sys_oid) = $cce->find('Vserver', {'vpsid' => $vmid});
    ($ok, $vps_data) = $cce->get($sys_oid);
    if ($ok) {
        $vps_fqdn		= $vps_data->{'hostname'};
        $vps_ip		= $vps_data->{'ip'};
    }
}
else {
    # Not OpenVZ - then it's KVM:
    ($sys_oid) = $cce->find('KVM', {'vpsid' => $vmid});
    ($ok, $vps_data) = $cce->get($sys_oid);
    if ($ok) {
        $vps_fqdn		= $vps_data->{'hostname'};
        $vps_ip		= $vps_data->{'ip'};
    }
}

# Fallbacks:
if (!$vps_fqdn) { $vps_fqdn = $hostname; }
if (!$vps_ip) { $vps_ip = "127.0.0.1"; }


#
##### Backup completed:
#

if ($phase eq 'backup-end') {


    # As local backup has just ended, we now create the CODB object that says so:
    ($ok) = $cce->create('Backup', {
        'origin' => 'vzdump',
        'vpsid' => $vmid,
        'virttype' => $virttype,
        'location' => 'local',
        'LBtimestamp' => $datetime,
        'storage' => $tarfile,
        'ip' => $vps_ip,
        'hostname' => $vps_fqdn,
        'timestamp' => time()
    });

    # If remote backup is enabled, copy the backup to the target destination:
    if ($user && $pass && $port && ($targethost||$targetip) && $path && ($remote_on == "1")) {
	if ($targethost) {
	    $target = $user . '@' . $targethost;
	}
	else {
	    $target = $user . '@' . $targetip;
	}

	# Connect to remote storage:
	no warnings 'Net::SFTP::Foreign';
	$sftp = Net::SFTP::Foreign->new($target, password => $pass, port => $port, autodie => 0, stderr_fh => $ssherr);
	$sftp->die_on_error("Unable to connect to remote host");

	# First get the filename of the tarball without the local parts of the path:
	$base_name = basename($tarfile);

	# Assemble target filename with path:
	$target_tarfile = $path . '/' . $base_name;

	# Transfer tarball to remote storage:
	$fh1 = $sftp->put($tarfile, $target_tarfile, late_set_perm => 1);
	# or die "Unable to transfer $tarfile to remote storage: " . $sftp->error;

	# As remote backup has just ended, we now create the CODB object that says so:
	($ok) = $cce->create('Backup', {
    	    'origin' => 'vzdump',
    	    'vpsid' => $vmid,
    	    'virttype' => $virttype,
    	    'location' => 'remote',
    	    'LBtimestamp' => $datetime,
    	    'storage' => $target_tarfile,
    	    'ip' => $vps_ip,
    	    'hostname' => $vps_fqdn,
    	    'timestamp' => time()
	});
    }

    # If local backup is disabled, we remove the tarball:
    if (($local_on eq "0") && ($tarfile ne "/") && ($tarfile)) {
	system("/bin/rm -f $tarfile");
    }

}

#
##### log-end phase:
#

if ($phase eq 'log-end') {

    # Remove the logfile as we don't need it:
    if (($logfile ne "/") && ($logfile)) {
	system("/bin/rm -f $logfile");
    }

    # Remove local CODB 'Backup' entries for which we no longer have tarballs 
    # because --maxfiles expired them:
    @my_local_backups = $cce->findx('Backup', {'vpsid' => $vmid, 'location' => 'local', 'origin' => 'vzdzump'});
    for $entry (@my_local_backups) {
        ($ok, $old_entry) = $cce->get($entry);
        $is_file = $old_entry->{'storage'};
        if (!-f "$is_file") {
	    ($ok) = $cce->destroy($entry);
	}
    }

    # Remove remote CODB 'Backup' tarballs if we have more 'remote' Backup objects than 
    # --maxfiles allows. First build an array with all 'remote' 'Backup' objects of this
    # VPSID:

    $num_of_remotes = "0";
    @my_remote_backups = $cce->findx('Backup', {'vpsid' => $vmid, 'location' => 'remote', 'origin' => 'vzdzump'});
    for $entry (@my_remote_backups) {
	$num_of_remotes++;
    }

    # Remove excess remote backups:
    if ($num_of_remotes > $my_maxfiles) {
	# We have more remote backups than we want to keep:
	for $entry (@my_remote_backups) {
        	($ok, $old_entry) = $cce->get($entry);
		push(@remote_stored, $old_entry->{'timestamp'});
	}
	@sorted = sort { $a <=> $b } @remote_stored;
	
	# Remove as many newest entries from @sorted as we have maxfiles set to:
	$z = $my_maxfiles;
	for ($z = $my_maxfiles; $z > 0; $z--) {
	    pop(@sorted);
	}
	
	# We now have a sorted array with the remote backups. Remove the oldest tarball(s):
	foreach $entry (@sorted) {
	    @my_single = $cce->find('Backup', {'timestamp' => $entry, 'location' => 'remote', 'origin' => 'vzdzump'});
    	    ($ok, $expire_entry) = $cce->get($my_single[0]);
	    $exp_path = $expire_entry->{'storage'};

	    # Assemble target login details:
	    if ($targethost) {
		$target = $user . '@' . $targethost;
	    }
	    else {
		$target = $user . '@' . $targetip;
	    }

	    # Connect to remote storage:
	    no warnings 'Net::SFTP::Foreign';
	    $sftp = Net::SFTP::Foreign->new($target, password => $pass, port => $port, autodie => 0, stderr_fh => $ssherr);
	    $sftp->die_on_error("Unable to connect to remote host");

	    # Does the old file exist?
	    $fh1 = $sftp->stat("$exp_path");
	    if ($fh1) {
		# Remove file:
		$fh2 = $sftp->remove("$exp_path");
		if ($fh2) {
		    # Remove CODB Object:
		    ($ok) = $cce->destroy($my_single[0]);
		}
	    }
	    else {
		# Remove CODB Object:
		($ok) = $cce->destroy($my_single[0]);
	    }
	}
    }
}

$cce->bye('SUCCESS');
exit 0;

